from google.protobuf.struct_pb2 import Struct
from google.protobuf.empty_pb2 import Empty
from perxis.common import common_pb2
from perxis.items import items_pb2, items_pb2_grpc
from perxis.references import references_pb2
from perxis.references import references_pb2_grpc
from perxis.models import Reference
from perxis.channel import GrpcChannel
from perxis.files import files_pb2, files_pb2_grpc


DEFAULT_PAGE_SIZE: int = 100
DEFAULT_PART_SIZE: int = 1024 * 5


class PerxisFilesWrapper:
    __stub: files_pb2_grpc.FilesStub

    def __init__(self, stub: files_pb2_grpc.FilesStub) -> None:
        self.__stub = stub

    async def start_upload(
        self, file_name: str, file_size: int
    ) -> files_pb2.StartUploadResponse:
        message = await self.__stub.StartUpload(
            files_pb2.StartUploadRequest(
                upload=files_pb2.MultipartUpload(
                    file=files_pb2.File(name=file_name, size=file_size)
                )
            )
        )
        return message

    async def complete_upload(
        self,
        file_id: str,
        upload_id: str,
        parts: list[str],
        part_size: int = DEFAULT_PART_SIZE,
    ) -> files_pb2.CompleteUploadResponse:
        message = await self.__stub.CompleteUpload(
            files_pb2.CompleteUploadRequest(
                upload=files_pb2.MultipartUpload(
                    file=files_pb2.File(id=file_id),
                    upload_id=upload_id,
                    part_size=part_size,
                    parts=[
                        files_pb2.CompletedPart(id=value, number=index)
                        for index, value in enumerate(parts, 1)
                    ],
                )
            )
        )
        return message


class PerxisReferencesWrapper:
    __references: references_pb2_grpc.ReferencesStub

    def __init__(self, references: references_pb2_grpc.ReferencesStub):
        self.__references = references

    async def get_references(
        self, references: list[Reference], space_id: str, env_id: str
    ) -> items_pb2.GetResponse:
        """Метод получения связей."""
        message = await self.__references.Get(
            references_pb2.GetRequest(
                space_id=space_id,
                env_id=env_id,
                references=[
                    references_pb2.Reference(id=ref.id, collection_id=ref.collection_id)
                    for ref in references
                ],
            )
        )
        return message


class PerxisItemsWrapper:
    __items: items_pb2_grpc.ItemsStub

    def __init__(self, items: items_pb2_grpc.ItemsStub):
        self.__items = items

    async def get(
        self, item_id: str, collection_id: str, space_id: str, env_id: str
    ) -> items_pb2.GetResponse:
        """Получение записи по идентификатору."""
        message = await self.__items.Get(
            items_pb2.GetRequest(
                item_id=item_id,
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
            )
        )
        return message

    async def create(
        self, data: Struct, collection_id: str, space_id: str, env_id: str
    ) -> items_pb2.CreateResponse:
        message = await self.__items.Create(
            items_pb2.CreateRequest(
                item=items_pb2.Item(
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                    data=data,
                )
            )
        )
        return message

    async def update(
        self, item_id: str, data: Struct, collection_id: str, space_id: str, env_id: str
    ) -> Empty:
        message = await self.__items.Update(
            items_pb2.UpdateRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                    data=data,
                )
            )
        )
        return message

    async def find(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
        deleted: bool = False,
        regular: bool = False,
        hidden: bool = False,
        templates: bool = False,
    ) -> items_pb2.FindResponse:
        """Метод поиска записей в коллекции."""
        result = await self.__items.Find(
            items_pb2.FindRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.FindOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by,
                        page_num=page_num,
                        page_size=page_size,
                        fields=fields or [],
                        limit=limit,
                        offset=offset,
                        exclude_fields=exclude_fields,
                    ),
                    deleted=deleted,
                    regular=regular,
                    hidden=hidden,
                    templates=templates,
                ),
            )
        )
        return message

    async def find_published(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        sort_by: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска опубликованных записей в коллекции."""
        message = await self.__items.FindPublished(
            items_pb2.FindPublishedRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.FindPublishedOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by,
                        fields=fields or [],
                        limit=limit,
                        offset=offset,
                        page_num=page_num,
                        page_size=page_size,
                        exclude_fields=exclude_fields,
                    )
                ),
            )
        )
        return message

    async def fetch_all_published(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        sort_by: list[str] | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска всех опубликованных записей в коллекции."""
        kwargs = {
            "collection_id": collection_id,
            "filters": filters,
            "sort_by": sort_by,
            "fields": fields,
            "exclude_fields": exclude_fields,
            "limit": page_size,
            "offset": 0,
            "space_id": space_id,
            "env_id": env_id,
        }
        message = await self.find_published(**kwargs)
        yield message

        if pages := message.total // page_size:
            if message.total % page_size:
                pages += 1
        else:
            pages = 1

        for page_num in range(2, pages + 1):
            offset = (page_num - 1) * page_size
            kwargs["offset"] = offset
            yield await self.find_published(**kwargs)

    async def unpublish(
        self, item_id: str, collection_id: str, space_id: str, env_id: str
    ) -> Empty:
        """Метод снятия с публикации записи в коллекции."""
        message = await self.__items.Unpublish(
            items_pb2.UnpublishRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                )
            )
        )
        return message

    async def publish(
        self, item_id: str, collection_id: str, space_id: str, env_id: str
    ) -> Empty:
        """Метод публикации записи в коллекции."""
        message = await self.__items.Publish(
            items_pb2.PublishRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                )
            )
        )
        return message

    async def fetch_all(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        sort_by: list[str] | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
        deleted: bool = False,
        regular: bool = False,
        hidden: bool = False,
        templates: bool = False,
    ) -> items_pb2.FindResponse:
        """Метод получения всех записей коллекции."""
        items = []
        storage_data = await self.find(
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            limit=page_size,
            offset=0,
            filters=filters,
            fields=fields,
            exclude_fields=exclude_fields,
            deleted=deleted,
            regular=regular,
            hidden=hidden,
            templates=templates,
        )
        items.extend(storage_data.items)

        if pages := storage_data.total // page_size:
            if storage_data.total % page_size:
                pages += 1
        else:
            pages = 1

        for page_num in range(2, pages + 1):
            offset = (page_num - 1) * page_size

            storage_data = await self.find(
                collection_id=collection_id,
                space_id=space_id,
                env_id=env_id,
                filters=filters,
                sort_by=sort_by,
                limit=page_size,
                offset=offset,
                fields=fields,
                exclude_fields=exclude_fields,
                deleted=deleted,
                regular=regular,
                hidden=hidden,
                templates=templates,
            )
            items.extend(storage_data.items)
        return items_pb2.FindResponse(items=items, total=len(items))

    async def delete(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        update_attrs: bool = False,
        erase: bool = True,
        **kwargs,
    ) -> Empty:
        message = await self.__items.Delete(
            items_pb2.DeleteRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                    **kwargs,
                ),
                options=items_pb2.DeleteOptions(update_attrs=update_attrs, erase=erase),
            )
        )
        return message

    async def undelete(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        update_attrs: bool = False,
        **kwargs,
    ) -> Empty:
        message = await self.__items.Undelete(
            items_pb2.UndeleteRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                    **kwargs,
                ),
                options=items_pb2.UndeleteOptions(update_attrs=update_attrs),
            )
        )
        return message

    async def get_published(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        locale_id: str | None = None,
    ) -> items_pb2.GetPublishedResponse:
        message = await self.__items.GetPublished(
            items_pb2.GetPublishedRequest(
                item_id=item_id,
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                options=items_pb2.GetPublishedOptions(locale_id=locale_id),
            )
        )
        return message

    async def aggregate(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        aggregate_options: dict[str, str] | None = None,
        filters: list[str] | None = None,
    ) -> items_pb2.AggregateResponse:
        message = await self.__items.Aggregate(
            items_pb2.AggregateRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.AggregateOptions(fields=aggregate_options),
            )
        )
        return message

    async def aggregate_published(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        aggregate_options: dict[str, str] | None = None,
        filters: list[str] | None = None,
    ) -> items_pb2.AggregatePublishedResponse:
        message = await self.__items.AggregatePublished(
            items_pb2.AggregatePublishedRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.AggregatePublishedOptions(fields=aggregate_options),
            )
        )
        return message

    async def get_revision(
        self,
        item_id: str,
        revision_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> items_pb2.GetRevisionResponse:
        message = await self.__items.GetRevision(
            items_pb2.GetRevisionRequest(
                item_id=item_id,
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                revision_id=revision_id,
            )
        )
        return message

    async def list_revisions(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
    ) -> items_pb2.ListRevisionsResponse:
        message = await self.__items.ListRevisions(
            items_pb2.ListRevisionsRequest(
                item_id=item_id,
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                options=items_pb2.ListRevisionsOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by or [],
                        fields=fields or [],
                        exclude_fields=exclude_fields,
                        offset=offset,
                        limit=limit,
                    )
                ),
            )
        )
        return message

    async def archive(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> Empty:
        message = await self.__items.Archive(
            items_pb2.ArchiveRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                )
            )
        )
        return message

    async def unarchive(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> Empty:
        message = await self.__items.Unarchive(
            items_pb2.UnarchiveRequest(
                item=items_pb2.Item(
                    id=item_id,
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=collection_id,
                )
            )
        )
        return message

    async def find_archived(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
    ) -> items_pb2.FindArchivedResponse:
        message = await self.__items.FindArchived(
            items_pb2.FindArchivedRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=collection_id,
                filter=items_pb2.Filter(q=filters or []),
                options=items_pb2.FindArchivedOptions(
                    options=common_pb2.FindOptions(
                        sort=sort_by or [],
                        fields=fields or [],
                        exclude_fields=exclude_fields,
                        offset=offset,
                        limit=limit,
                    )
                ),
            )
        )
        return message


class PerxisDataProvider:
    def __init__(
        self,
        channel: GrpcChannel,
        space_id: str,
        env_id: str,
    ) -> None:
        self.channel = channel
        self.space_id = space_id
        self.env_id = env_id
        self.references_wrapper = PerxisReferencesWrapper(
            references=references_pb2_grpc.ReferencesStub(self.channel.channel)
        )
        self.items_wrapper = PerxisItemsWrapper(
            items=items_pb2_grpc.ItemsStub(self.channel.channel),
        )

    async def get(self, item_id: str, collection_id: str) -> items_pb2.GetResponse:
        """Получение записи по идентификатору."""
        message = await self.items_wrapper.get(
            item_id=item_id,
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
        )
        return message

    async def create(
        self, data: Struct, collection_id: str
    ) -> items_pb2.CreateResponse:
        message = await self.items_wrapper.create(
            data=data,
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
        )
        return message

    async def update(self, item_id: str, data: Struct, collection_id: str) -> Empty:
        message = await self.items_wrapper.update(
            item_id=item_id,
            data=data,
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
        )
        return message

    async def find(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
        deleted: bool = False,
        regular: bool = False,
        hidden: bool = False,
        templates: bool = False,
    ) -> items_pb2.FindResponse:
        message = await self.items_wrapper.find(
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
            filters=filters,
            sort_by=sort_by,
            fields=fields,
            exclude_fields=exclude_fields,
            limit=limit,
            offset=offset,
            page_num=page_num,
            page_size=page_size,
            deleted=deleted,
            regular=regular,
            hidden=hidden,
            templates=templates,
        )
        return message

    async def find_published(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        sort_by: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
        page_num: int | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        message = await self.items_wrapper.find_published(
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
            filters=filters,
            fields=fields,
            exclude_fields=exclude_fields,
            sort_by=sort_by,
            limit=limit,
            offset=offset,
            page_num=page_num,
            page_size=page_size,
        )
        return message

    async def fetch_all_published(
        self,
        collection_id: str,
        filters: list[str] | None = None,
        fields: list[str] | None = None,
        sort_by: list[str] | None = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        """Метод поиска всех опубликованных записей в коллекции."""
        kwargs = {
            "collection_id": collection_id,
            "filters": filters,
            "sort_by": sort_by,
            "fields": fields,
            "limit": page_size,
            "offset": 0,
        }
        message = await self.find_published(**kwargs)
        yield message

        if pages := message.total // page_size:
            if message.total % page_size:
                pages += 1
        else:
            pages = 1

        for page_num in range(2, pages + 1):
            offset = (page_num - 1) * page_size
            kwargs["offset"] = offset
            yield await self.find_published(**kwargs)

    async def unpublish(self, item_id: str, collection_id: str) -> Empty:
        message = await self.items_wrapper.unpublish(
            item_id=item_id,
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
        )
        return message

    async def publish(self, item_id: str, collection_id: str) -> Empty:
        message = await self.items_wrapper.publish(
            item_id=item_id,
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
        )
        return message

    async def fetch_all(
        self,
        collection_id: str,
        filters: list[str] | str = None,
        sort_by: list[str] | str = None,
        page_size: int | None = DEFAULT_PAGE_SIZE,
    ) -> items_pb2.FindResponse:
        message = await self.items_wrapper.fetch_all(
            collection_id=collection_id,
            space_id=self.space_id,
            env_id=self.env_id,
            filters=filters,
            sort_by=sort_by,
            page_size=page_size,
        )
        return message

    async def get_references(
        self, references: list[Reference]
    ) -> items_pb2.GetResponse:
        message = await self.references_wrapper.get_references(
            references=references, space_id=self.space_id, env_id=self.env_id
        )
        return message

    async def delete(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        update_attrs: bool = False,
        erase: bool = True,
        **kwargs,
    ) -> Empty:
        message = await self.items_wrapper.delete(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            update_attrs=update_attrs,
            erase=erase,
            **kwargs,
        )
        return message

    async def undelete(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        update_attrs: bool = False,
        **kwargs,
    ) -> Empty:
        message = await self.items_wrapper.undelete(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            update_attrs=update_attrs,
            **kwargs,
        )
        return message

    async def get_published(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        locale_id: str | None = None,
    ) -> items_pb2.GetPublishedResponse:
        message = await self.items_wrapper.get_published(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            locale_id=locale_id,
        )
        return message

    async def aggregate(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        aggregate_options: dict[str, str] | None = None,
        filters: list[str] | None = None,
    ) -> items_pb2.AggregateResponse:
        message = await self.items_wrapper.aggregate(
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            aggregate_options=aggregate_options,
            filters=filters,
        )
        return message

    async def aggregate_published(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        aggregate_options: dict[str, str] | None = None,
        filters: list[str] | None = None,
    ) -> items_pb2.AggregatePublishedResponse:
        message = await self.items_wrapper.aggregate_published(
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            aggregate_options=aggregate_options,
            filters=filters,
        )
        return message

    async def get_revision(
        self,
        item_id: str,
        revision_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> items_pb2.GetRevisionResponse:
        message = await self.items_wrapper.get_revision(
            item_id=item_id,
            revision_id=revision_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
        )
        return message

    async def list_revisions(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
    ) -> items_pb2.ListRevisionsResponse:
        message = await self.items_wrapper.list_revisions(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            sort_by=sort_by,
            fields=fields,
            exclude_fields=exclude_fields,
            limit=limit,
            offset=offset,
        )
        return message

    async def archive(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> Empty:
        message = await self.items_wrapper.archive(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
        )
        return message

    async def unarchive(
        self,
        item_id: str,
        collection_id: str,
        space_id: str,
        env_id: str,
    ) -> Empty:
        message = await self.items_wrapper.archive(
            item_id=item_id,
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
        )
        return message

    async def find_archived(
        self,
        collection_id: str,
        space_id: str,
        env_id: str,
        filters: list[str] | None = None,
        sort_by: list[str] | None = None,
        fields: list[str] | None = None,
        exclude_fields: bool = False,
        limit: int | None = None,
        offset: int | None = None,
    ) -> items_pb2.FindArchivedResponse:
        message = await self.items_wrapper.find_archived(
            collection_id=collection_id,
            space_id=space_id,
            env_id=env_id,
            filters=filters,
            sort_by=sort_by,
            fields=fields,
            exclude_fields=exclude_fields,
            limit=limit,
            offset=offset,
        )
        return message


class PerxisFileProvider:
    __files_wrapper: PerxisFilesWrapper

    def __init__(self, channel: GrpcChannel) -> None:
        self.__files_wrapper = PerxisFilesWrapper(
            stub=files_pb2_grpc.FilesStub(channel.channel)
        )

    async def start_upload(
        self, file_name: str, file_size: int
    ) -> files_pb2.StartUploadResponse:
        message = await self.__files_wrapper.start_upload(
            file_name=file_name, file_size=file_size
        )
        return message

    async def complete_upload(
        self,
        file_id: str,
        upload_id: str,
        parts: list[str],
        part_size: int = DEFAULT_PART_SIZE,
    ) -> files_pb2.CompleteUploadResponse:
        message = await self.__files_wrapper.complete_upload(
            file_id=file_id, upload_id=upload_id, parts=parts, part_size=part_size
        )
        return message
